import logging
from typing import Optional, Tuple, Union

import torch
import torch.nn as nn
from torch import Tensor

from mae import MAE
from utils import get_checkpoint_path_without_suffix

logger = logging.getLogger(name=__name__)


def get_net_architecture(
    in_features: int,
    out_features: int,
    net_architecture: str = "mlp",
    dropout: float = 0.0,
) -> nn.Module:
    if net_architecture == "mlp":
        net = nn.Sequential(
            nn.Linear(in_features=in_features, out_features=128),
            nn.ReLU(),
            nn.BatchNorm1d(num_features=128),
            nn.Dropout(dropout),
            nn.Linear(in_features=128, out_features=16),
            nn.ReLU(),
            nn.BatchNorm1d(num_features=16),
            nn.Dropout(dropout),
            nn.Linear(in_features=16, out_features=out_features),
        )
    else:
        raise ValueError

    return net


class Classifier(nn.Module):
    """Classifier initialized with pretrained T-MAE, here with finetuning/training of T-MAE parameters."""

    def __init__(
        self,
        mae: MAE,  # assumes a restored pretrained T-MAE!
        num_classes: int = 6,
        class_weights: Optional[Tensor] = None,
        net_architecture: str = "mlp",
        dropout: float = 0.3,
        restore_from: Optional[str] = None,
        restore_iteration: Union[str, int] = "best",
    ) -> None:
        super().__init__()
        logger.info("Initializing model ...")

        self.mae = mae
        self.num_classes = num_classes

        in_features = self.mae.num_patches * self.mae.embed_dim

        self.net = get_net_architecture(
            in_features=in_features,
            out_features=self.num_classes,
            net_architecture=net_architecture,
            dropout=dropout,
        )

        self.loss_fn = nn.BCEWithLogitsLoss(reduction="mean", pos_weight=class_weights)

        if restore_from is not None:
            self.restore(restore_dir=restore_from, iteration=restore_iteration)

    def forward(self, x: Tensor, target: Tensor) -> Tuple[Tensor, Tensor]:
        logits = self.forward_logits(x)

        loss = self.loss_fn(logits, target.type(torch.float32))
        preds = torch.sigmoid(logits)

        return loss, preds

    def forward_logits(self, x: Tensor) -> Tensor:
        B, t, N, f = x.shape
        assert (t, N, f) == (
            self.mae.seq_len,
            self.mae.num_agents,
            self.mae.num_features,
        )
        latent, _, ids_restore = self.mae.forward_encoder(x, mask_ratio=0.0)
        # unshuffle
        B_, p, N_, d = latent.shape
        assert ids_restore.shape[1] == self.mae.num_patches
        latent = torch.gather(
            latent, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, 1, d)
        )

        B_, p, N_, d = latent.shape
        assert (B, N) == (B_, N_)
        assert (p, d) == (
            self.mae.num_patches,
            self.mae.embed_dim,
        )

        input = torch.einsum("BpNd->Bpd", latent).reshape(B, p * d)

        output = self.net(input)

        return output

    def restore(
        self,
        restore_dir: str,
        iteration: Union[str, int] = "best",
    ) -> None:
        logger.info(f"Restoring model from {restore_dir}")
        checkpoint_path_without_suffix = get_checkpoint_path_without_suffix(
            dir=restore_dir, iteration=iteration
        )
        checkpoint: dict = torch.load(
            checkpoint_path_without_suffix.with_suffix(".pth"),
            map_location=torch.device("cpu"),
        )
        self.load_state_dict(checkpoint.get("model_state_dict"))
