import logging
import math
from typing import Optional, Tuple, Union

import torch
import torch.nn as nn
from torch import Tensor

from utils import get_checkpoint_path_without_suffix

logger = logging.getLogger(name=__name__)


class PositionalEncoding(nn.Module):
    """Adapted from https://pytorch.org/tutorials/beginner/transformer_tutorial.html"""

    def __init__(self, d_model: int, dropout: float = 0.0, max_len: int = 100):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model)
        )
        pe = torch.zeros(max_len, d_model)
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).unsqueeze(2)  # adjust shape to data shape
        self.register_buffer("pe", pe)

    def forward(self, x: Tensor) -> Tensor:
        B, num_patches, num_agents, embed_dim = x.shape
        x = x + self.pe[:, :num_patches]
        return self.dropout(x)


class MAE(nn.Module):
    """MAE generalized to multiagent trajectory data.

    Based in parts on MAE code (https://github.com/facebookresearch/mae)
    and "autobots" code for encoder and decoder layers (https://github.com/roggirg/AutoBots).

    See our paper for more references."""

    def __init__(
        self,
        seq_len: int,
        num_agents: int,
        num_features: int,
        patch_len: int,
        embed_dim: int = 256,
        num_layers: int = 4,
        num_heads: int = 8,
        dropout: float = 0.0,
        dim_feedforward: int = 512,
        decoder_embed_dim: int = 128,
        decoder_num_layers: int = 4,
        decoder_num_heads: int = 8,
        decoder_dropout: float = 0.0,
        decoder_dim_feedforward: int = 512,
        restore_from: Optional[str] = None,
        restore_iteration: Union[str, int] = "best",
    ) -> None:
        super().__init__()
        self.seq_len = seq_len
        self.num_agents = num_agents
        self.num_features = num_features
        self.patch_len = patch_len

        num_patches = self.seq_len / patch_len
        # assert that num_patch is integer and cast type
        assert int(num_patches) + 0.0 == num_patches
        self.num_patches = int(num_patches)

        ### network parts ###

        # encoder
        self.embed_dim = embed_dim
        self.num_layers = num_layers
        self.num_heads = num_heads
        self.dropout = dropout
        self.dim_feedforward = dim_feedforward
        self.embed = nn.Linear(
            in_features=self.patch_len * self.num_features,
            out_features=self.embed_dim,
        )
        self.pos_embed = PositionalEncoding(
            d_model=self.embed_dim, max_len=self.num_patches
        )

        # autobots encoder
        self.social_attn_layers = []
        self.temporal_attn_layers = []
        for _ in range(self.num_layers):
            tx_encoder_layer = nn.TransformerEncoderLayer(
                d_model=self.embed_dim,
                nhead=self.num_heads,
                dropout=self.dropout,
                dim_feedforward=self.dim_feedforward,
                batch_first=False,
            )
            self.temporal_attn_layers.append(
                nn.TransformerEncoder(
                    tx_encoder_layer, num_layers=1
                )  # (set to 1 layer for now to have no duplication)
            )
            tx_encoder_layer = nn.TransformerEncoderLayer(
                d_model=self.embed_dim,
                nhead=self.num_heads,
                dropout=self.dropout,
                dim_feedforward=self.dim_feedforward,
                batch_first=False,
            )
            self.social_attn_layers.append(
                nn.TransformerEncoder(tx_encoder_layer, num_layers=1)
            )

        self.temporal_attn_layers = nn.ModuleList(self.temporal_attn_layers)
        self.social_attn_layers = nn.ModuleList(self.social_attn_layers)

        # decoder
        self.decoder_embed_dim = decoder_embed_dim
        self.decoder_num_layers = decoder_num_layers
        self.decoder_num_heads = decoder_num_heads
        self.decoder_dropout = decoder_dropout
        self.decoder_dim_feedforward = decoder_dim_feedforward
        self.decoder_embed = nn.Linear(self.embed_dim, self.decoder_embed_dim)
        self.mask_token = nn.Parameter(torch.zeros(1, 1, 1, self.decoder_embed_dim))
        self.decoder_pos_embed = PositionalEncoding(
            d_model=self.decoder_embed_dim, max_len=self.num_patches
        )
        self.decoder_pred = nn.Linear(
            self.decoder_embed_dim, self.patch_len * self.num_features
        )  # decoder to patch

        # autobots decoder
        self.social_attn_decoder_layers = []
        self.temporal_attn_decoder_layers = []
        for _ in range(self.decoder_num_layers):
            tx_decoder_layer = nn.TransformerEncoderLayer(
                d_model=self.decoder_embed_dim,
                nhead=self.decoder_num_heads,
                dropout=self.decoder_dropout,
                dim_feedforward=self.decoder_dim_feedforward,
                batch_first=False,
            )
            self.temporal_attn_decoder_layers.append(
                nn.TransformerEncoder(
                    tx_decoder_layer, num_layers=1
                )  # TODO: (set to 1 layer for now to have no duplication)
            )
            tx_encoder_layer = nn.TransformerEncoderLayer(
                d_model=self.decoder_embed_dim,
                nhead=self.decoder_num_heads,
                dropout=self.decoder_dropout,
                dim_feedforward=self.decoder_dim_feedforward,
                batch_first=False,
            )
            self.social_attn_decoder_layers.append(
                nn.TransformerEncoder(tx_encoder_layer, num_layers=1)
            )

        self.temporal_attn_decoder_layers = nn.ModuleList(
            self.temporal_attn_decoder_layers
        )
        self.social_attn_decoder_layers = nn.ModuleList(self.social_attn_decoder_layers)

        self.initialize_weights()

        if restore_from is not None:
            self.restore(restore_dir=restore_from, iteration=restore_iteration)

    def initialize_weights(self) -> None:
        # TODO: Refactor for better init weights of autobots transformer layers (IMPORTANT only if we allow layer duplication!)

        # init mask_token
        torch.nn.init.normal_(self.mask_token, std=0.02)

        # init all linear and norm layers
        self.apply(self._init_linear_and_norm_weights)

    def _init_linear_and_norm_weights(self, m) -> None:
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def patchify(self, x: Tensor) -> Tensor:
        B, S, N, f = x.shape
        assert (S, N, f) == (self.seq_len, self.num_agents, self.num_features)
        x = x.reshape(B, self.num_patches, self.patch_len, N, f)
        x = torch.einsum("BplNf->BpNlf", x)
        x = x.reshape(B, self.num_patches, N, self.patch_len * f)
        return x

    def unpatchify(self, x: Tensor) -> Tensor:
        B, p, N, lf = x.shape
        assert (p, N) == (self.num_patches, self.num_agents)
        x = x.reshape(B, self.num_patches, N, self.patch_len, self.num_features)
        x = torch.einsum("BpNlf->BplNf", x)
        x = x.reshape(B, self.num_patches * self.patch_len, N, self.num_features)
        return x

    def random_masking(
        self, x: Tensor, mask_ratio: float
    ) -> Tuple[Tensor, Tensor, Tensor]:
        B, p, N, d = x.shape
        assert (p, N) == (self.num_patches, self.num_agents)
        len_keep = round(p * (1 - mask_ratio))

        noise = torch.rand(B, p, N, device=x.device)
        ids_shuffle = torch.argsort(
            noise, dim=1
        )  # ascend: small is keep, large is remove
        ids_restore = torch.argsort(ids_shuffle, dim=1)  # yr: ids to revert shuffle

        # keep the first subset
        ids_keep = ids_shuffle[:, :len_keep]
        x_masked = torch.gather(
            x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, 1, d)
        )

        # generate the binary mask: 0 is keep, 1 is remove
        mask = torch.ones([B, p, N], device=x.device)
        mask[:, :len_keep] = 0
        # unshuffle to get the binary mask
        mask = torch.gather(mask, dim=1, index=ids_restore)

        return x_masked, mask, ids_restore

    def forward_encoder(
        self, x: Tensor, mask_ratio: float
    ) -> Tuple[Tensor, Tensor, Tensor]:

        # to input tokens
        x = self.patchify(x)
        x = self.embed.forward(x)
        B, p, N, d = x.shape

        # add pos embed
        x = self.pos_embed.forward(x)

        # masking: length -> length * mask_ratio
        x, mask, ids_restore = self.random_masking(x, mask_ratio)

        # run through autobot encoder
        x = self.autobots_encoder(x=x)

        return x, mask, ids_restore

    def autobots_encoder(self, x: Tensor) -> Tensor:
        B, p, N, d = x.shape
        for i in range(self.num_layers):
            x = torch.einsum("BpNd->pBNd", x).reshape(p, B * N, d)
            x = self.temporal_attn_layers[i](x)
            assert x.shape == (p, B * N, d)
            x = torch.einsum("pBNd->NBpd", x.reshape(p, B, N, d)).reshape(N, B * p, d)
            x = self.social_attn_layers[i](x)
            assert x.shape == (N, B * p, d)
            x = torch.einsum("NBpd->BpNd", x.reshape(N, B, p, d))
            assert x.shape == (B, p, N, d)
        return x

    def forward_decoder(self, x: Tensor, ids_restore: Tensor) -> Tensor:
        # embed tokens
        x = self.decoder_embed(x)
        B, p, N, d = x.shape

        # append mask tokens to sequence
        assert ids_restore.shape[1] == self.num_patches
        mask_tokens = self.mask_token.repeat(B, self.num_patches - p, N, 1)
        x_ = torch.cat([x, mask_tokens], dim=1)
        x_ = torch.gather(
            x_, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, 1, d)
        )  # unshuffle
        x = x_

        # add pos embed
        x = self.decoder_pos_embed(x)

        # run through autobot decoder
        x = self.autobots_decoder(x)

        # predictor projection
        x = self.decoder_pred(x)

        return x

    def autobots_decoder(self, x: Tensor) -> Tensor:
        B, p, N, d = x.shape
        for i in range(self.decoder_num_layers):
            x = torch.einsum("BpNd->pBNd", x).reshape(p, B * N, d)
            x = self.temporal_attn_decoder_layers[i](x)
            assert x.shape == (p, B * N, d)
            x = torch.einsum("pBNd->NBpd", x.reshape(p, B, N, d)).reshape(N, B * p, d)
            x = self.social_attn_decoder_layers[i](x)
            assert x.shape == (N, B * p, d)
            x = torch.einsum("NBpd->BpNd", x.reshape(N, B, p, d))
            assert x.shape == (B, p, N, d)
        return x

    def forward_loss(
        self,
        target: Tensor,
        pred: Tensor,
        mask: Tensor,
        only_loss_for_masked_part: bool = True,
    ) -> Tensor:
        target = self.patchify(target)
        assert target.shape == pred.shape

        loss = (pred - target) ** 2
        loss = loss.mean(dim=-1)  # [B, p, N], mean loss per patch

        if only_loss_for_masked_part:
            loss = (loss * mask).sum() / mask.sum()  # mean loss on removed patches
        else:
            loss = (
                loss.sum() / torch.ones_like(mask).sum()
            )  # loss on everything (only for inspection purposes)
        return loss

    def forward(
        self, x: Tensor, mask_ratio: float, only_loss_for_masked_part: bool = True
    ) -> Tuple[Tensor, Tensor, Tensor]:
        latent, mask, ids_restore = self.forward_encoder(x=x, mask_ratio=mask_ratio)
        pred = self.forward_decoder(x=latent, ids_restore=ids_restore)
        loss = self.forward_loss(
            target=x,
            pred=pred,
            mask=mask,
            only_loss_for_masked_part=only_loss_for_masked_part,
        )  # loss is mean loss per unmasked patches

        # reshaping and updating prediction
        pred_in_dataspace = self.unpatchify(pred)
        assert pred_in_dataspace.shape == x.shape

        x_patchified = self.patchify(x)
        assert pred.shape == x_patchified.shape
        # merge kept x with pred via mask (0 is keep, 1 is remove)
        pred_incl_kept_x = (pred * mask.unsqueeze(-1)) + (
            x_patchified * (~mask.unsqueeze(-1).type(torch.bool)).type(torch.float32)
        )
        pred_incl_kept_x_in_dataspace = self.unpatchify(pred_incl_kept_x)
        assert pred_incl_kept_x_in_dataspace.shape == x.shape

        return loss, pred_in_dataspace, pred_incl_kept_x_in_dataspace, mask

    def restore(
        self,
        restore_dir: str,
        iteration: Union[str, int] = "best",
    ) -> None:
        logger.info(f"Restoring model from {restore_dir}")
        checkpoint_path_without_suffix = get_checkpoint_path_without_suffix(
            dir=restore_dir, iteration=iteration
        )
        checkpoint: dict = torch.load(
            checkpoint_path_without_suffix.with_suffix(".pth"),
            map_location=torch.device("cpu"),
        )
        self.load_state_dict(checkpoint.get("model_state_dict"))
